/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July 2002 
  */
 
package lotos;

import lotos.*; 
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

/**
 * ImageButton - A button component with an image in it
 * Usage Functions
 * 1. Constructor           public ImageButton( Image image, int wid, int height );
 * 2. set state handle      public void addListener( imageButtonListener listener );
 * 3. enable the button     public void enable();
 * 4. disable the button    public void disable();
 * 5. toggle the state      public void toggleButton(boolean armed)
 *
 */

public class ImageButton extends Canvas {

    public static final int UNARMED = 0;
    public static final int ARMED = 1;
    public static final int DISABLED = 2;

    private static final Border defaultUnarmedBorder =
        new DefaultImageButtonBorder( UNARMED );
    private static final Border defaultArmedBorder =
        new DefaultImageButtonBorder( ARMED );
    private static final Border defaultDisabledBorder =
        new DefaultImageButtonBorder( DISABLED );

    private MediaTracker tracker;

    private Image images[] = new Image[3];
    private Border borders[] = new Border[3];

    private boolean generatedDisabled = false;
    private int buttonState = UNARMED;
    private imageButtonListener buttonListener;
    private ImageButton buttonSelf;
    private int displayWidth;
    private int displayHeight;

    /**
     * Constructs an ImageButton
     */
    public ImageButton() {
        tracker = new MediaTracker( this );
        setUnarmedBorder( defaultUnarmedBorder );
        setArmedBorder( defaultArmedBorder );
        setDisabledBorder( defaultDisabledBorder );
    }

    /**
     * Constructs an ImageButton with the given image.
     *
     * @param image         the image for all states of the button
     *                      (until other images are assigned)
     * @param wid           the displayed width of the image
     * @param hight         the displayed height of the image
     */
    public ImageButton( Image image, int wid, int height ) {
        this();
        setUnarmedImage( image );
        displayWidth = wid;
        displayHeight = height;
    }

    /**
     * Used internally to add the Image to the array and the MediaTracker,
     * start loading the image if necessary via the tracker's "checkID", and
     * repaint if necessary.
     *
     * @param id        the buttonState id (also used as image id for the MediaTracker)
     * @param image     the image, which is not supposed to be null
     */

	private synchronized void setImage( int id, Image image ) {
	    if ( images[id] != image ) {
       	    images[id] = image;
        	if ( image != null ) {
    	        tracker.addImage( image, id );
        	    tracker.checkID( id, true );
        	}
    	    if ( buttonState == id ) {
        	    repaint();
        	}
        }
	}

    /**
     * Sets the image to display when the button is not pressed or hilited
     * because of a mouse-over.  This image is also used in those other cases
     * if no alternative image is requested.
     *
     * @param image     the unarmed image
     */
	public void setUnarmedImage( Image image ) {
        setImage( UNARMED, image );
        if ( images[ARMED] == null ) {
            setArmedImage( image );
        }
        if ( ( images[DISABLED] == null ) ||
             generatedDisabled ) {
            setDisabledImage( null );
        }
	}

    /**
     * Sets the image to display when the button is pressed and the mouse
     * is still over the button.
     *
     * @param image     the armed image
     */
	public void setArmedImage( Image image ) {
   	    if ( image != null ) {
       	    setImage( ARMED, image );
   	    }
   	    else {
   	        setImage( ARMED, images[UNARMED] );
   	    }
	}

    /**
     * Sets the image to display when the button is disabled.
     *
     * @param image     the disabled image
     */
	public void setDisabledImage( Image image ) {
        generatedDisabled = false;
   	    if ( ( image == null ) &&
   	         ( images[UNARMED] != null ) ) {
            generatedDisabled = true;
            image = createImage( new FilteredImageSource( images[UNARMED].getSource(),
                                                          new DisableImageFilter() ) );
   	    }
	    setImage( DISABLED, image );
	}

    /**
     * Used internally to add the Border to the array and repaint if necessary.
     *
     * @param   id      the buttonState, used to index the array
     * @param   border  the Border, which is not supposed to be null
     */
	private synchronized void setBorder( int id, Border border ) {
	    if ( borders[id] != border ) {
       	    borders[id] = border;
       	    if ( buttonState == id ) {
       	        repaint();
       	    }
        }
	}

    /**
     * Sets the border to display when the button is not pressed or hilited
     * because of a mouse-over.  This border is also used in those other cases
     * if no alternative border is requested.
     *
     * @param border     the unarmed border
     */
	public void setUnarmedBorder( Border border ) {
        setBorder( UNARMED, border );
        if ( borders[ARMED] == null ) {
            setArmedBorder( border );
        }
        if ( borders[DISABLED] == null ) {
            setDisabledBorder( border );
        }
	}

    /**
     * Sets the border to display when the button is pressed and the mouse
     * is still over the button.
     *
     * @param border     the armed border
     */
	public void setArmedBorder( Border border ) {
   	    if ( border != null ) {
       	    setBorder( ARMED, border );
   	    }
   	    else {
   	        setBorder( ARMED, borders[UNARMED] );
   	    }
	}

    /**
     * Sets the border to display when the button is disabled.
     *
     * @param border     the disabled border
    */

   	public void setDisabledBorder( Border border ) {
   	    if ( border != null ) {
       	    setBorder( DISABLED, border );
   	    }
   	    else {
   	        setBorder( DISABLED, borders[UNARMED] );
   	    }
	    if ( buttonState == DISABLED ) {
    	    repaint();
    	}
	}

    /**
     * Gets the current buttonState id for the button
     *
     * @return     the button state integer id
     */
	public int getButtonState() {
	    return buttonState;
	}

    public boolean isArmed(){
        return buttonState == this.ARMED;
    }
    /**
     * Sets the current buttonState id for the button
     *
     * @param   buttonState     the button state integer id
     */
	protected void setButtonState( int buttonState ) {
	    if ( buttonState != this.buttonState ) {
    		this.buttonState = buttonState;
            repaint();
    	}
	}

    /**
     * Overrides awt.Component.disable() to also set the button state.
     */
    public void disableButton() {
        setButtonState( DISABLED );
        super.setEnabled(false);
    }

    /**
     * Overrides awt.Component.enable() to also set the button state.
     */
    public void enableButton() {
        setButtonState( UNARMED );
        super.setEnabled(true);
    }

    /**
     * toggle the state of the button, only when it is enabled
     *
     */
     public void toggleButton(boolean armed){
        if( this.isEnabled() == false ) return;

        if( armed )
            setButtonState( ARMED );
        else
            setButtonState( UNARMED );
    }

    /**
     * Overrides awt.Component.paint() to paint the current border and image.
     *
     * @param     g     The Graphics in which to draw
     */
    public void paint( Graphics g ) {
        Dimension size = getSize();
        borders[buttonState].paint( g, getBackground(), 0, 0, size.width, size.height );
        try {
            if ( ! tracker.checkID( buttonState ) ) {
                tracker.waitForID( buttonState );
            }
            if ( ! tracker.isErrorID( buttonState ) ) {
                Insets insets = borders[buttonState].getInsets();
                int imageWidth = displayWidth; //images[buttonState].getWidth( this );
                int imageHeight = displayHeight; //images[buttonState].getHeight( this );
                int x = insets.left +
                        ( ( ( size.width - ( insets.left + insets.right ) ) -
                            imageWidth ) / 2 );
                int y = insets.top +
                        ( ( ( size.height - ( insets.top + insets.bottom ) ) -
                            imageHeight ) / 2 );
                g.drawImage( images[buttonState], x, y, displayWidth, displayHeight, this );
            }
        }
        catch ( InterruptedException ie ) {
        }
    }

    /**
     * Overrides awt.Component.preferredSize() to return the preferred size of the button.
     * This assumes the images (if more than one) are all the same size.  It also calculates
     * the maximum insets from all borders and adds them to the image dimensions.
     *
     * @param     g     The Graphics in which to draw
     */
    public Dimension preferredSize() {
        Dimension pref = new Dimension();
        try {
            if ( ! tracker.checkID( buttonState ) ) {
                tracker.waitForID( buttonState );
            }
            if ( ! tracker.isErrorID( buttonState ) ) {
                Dimension size = getSize();
                //pref.width = images[buttonState].getWidth( this );
                pref.width = displayWidth;
                //pref.height = images[buttonState].getHeight( this );
                pref.height = displayWidth;
            }
            int maxWidthAdd = 0;
            int maxHeightAdd = 0;
            for ( int i = 0; i < DISABLED; i++ ) {
                Insets insets = borders[i].getInsets();
                maxWidthAdd = Math.max( maxWidthAdd, insets.left+insets.right );
                maxHeightAdd = Math.max( maxHeightAdd, insets.top+insets.bottom );
            }
            pref.width += maxWidthAdd;
            pref.height += maxHeightAdd;
        }
        catch ( InterruptedException ie ) {
        }
        return pref;
    }

    /**
     * Overrides awt.Component.mouseUp() and invokes "action" if the button was ARMED
     * (because the mouse was over the button when it was released).
     *
     * @param   evt     The mouse up event
     * @param   x       The mouse x position
     * @param   y       The mouse y position
     * @return  true if the event was handled
     */
    /*public boolean mouseUp( Event evt, int x, int y ) {

        if ( inside( x, y ) ) {

            if ( ! action( evt, evt.arg ) ) {
                Container parent = getParent();
                while ( ( parent != null ) &&
                        ( ! parent.action( evt, evt.arg ) ) ) {
                    parent = parent.getParent();
                }
            }
        }
        return true;
    }*/

    /**
     * set the state change listener
     * @param listener      the recipient of the action
     */
    public void addListener( imageButtonListener listener ){
        buttonListener = listener;
        buttonSelf = this;
        addMouseListener(
            new MouseAdapter() {
                public void mouseClicked( MouseEvent me ){
                    /*if( buttonState == ARMED ) {
                        buttonState = UNARMED;
                    }else if( buttonState == UNARMED ){
                        buttonState = ARMED;
                    }*/
                    //change the state only when necessary
                    buttonListener.buttonStateChanged( buttonSelf );
                    repaint();
                }
            }
        );
    }

}

/**
 * DisableImageFilter - an RGBImageFilter that "greys out" an image by "blanking out"
 * every other pixel.
 */
class DisableImageFilter extends RGBImageFilter
{
    /**
     * Constructs a DisableImageFilter.  The canFilterIndexColorModel is set to false
     * because the pixel index is important during filtering.
     */
    public DisableImageFilter() {
        canFilterIndexColorModel = false;
    }

    /**
     * Called when a disabled image is created to alter the pixels to be blanked out.
     *
     * @param   x   the x position of the pixel
     * @param   y   the y position of the pixel
     * @param   rgb the rgb value of the pixel
     */
    public int filterRGB( int x, int y, int rgb ) {
        if ( ( ( x % 2 ) ^ ( y % 2 ) ) == 1 ) {
            return ( rgb & 0xffffff );
        }
        else {
            return rgb;
        }
    }


}

   /** the default button classes
     *
     */
    class DefaultImageButtonBorder extends Border {

        public DefaultImageButtonBorder( int state ) {
            setBorderThickness( 2 );
            switch(state) {
            case ImageButton.ARMED:
                setType( THREED_IN );
                setMargins( 4, 4, 2, 2 );
                break;
            case ImageButton.UNARMED:
                setType( THREED_OUT );
                setMargins( 3 );
                break;
            case ImageButton.DISABLED:
                setType( EMBOSSED_OUT );
                setMargins( 3 );
            }
        }
    }
