/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.util.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.io.*;
import hypercast.*;

/**
 * Overlay Network, the main class for the simulation.
 * Main components
 * a. physical nodes </p>
 *      the physical node </p>
 *      the protocol </p>
 *      the physical links and weights </p>
 *      the logical links </p>
 * b. a set of the active nodes within the network </p>
 * c. a discrete event simulator </p>
 * Main functions </p>
 * a. physical structure maintenence ( nodes add/deletion, physical link add/deletion) </p>
 * b. import/export physical topology </p>
 * c. import/export topology with protocol specific information </p>
 * d. execute the simulation, i.e. start, run, pause, step </p>
 * e. logical routing, used by the statisic collection package </p>
 * f. physical routing, used to provide services for the protocols </p>
 * g. graph-based simulation </p>
 *
 */
/**
 * revised by Guangyu Dong, Jan 6, 2004
 * to add mobility and wireless support
 */

abstract class OverlayNetwork
{
	
	long sum_sent=0;
	long sum_collision=0;
	long packet_sent=0;
	long packet_received=0;
	long bytes_sent=0;
	long bytes_received=0;
	
    final static Color phyNodeColor = Color.black;
    final static Color phyEdgeColor = Color.black;
	 final static float SIM_STEP = 0.005f;

	 /** a vector that holds all the nodes in the network */
    Vector nodes;
    /** a vector that holds all the nodes in the overlay network */
    Vector active_nodes;
    /** indicate whether a re-calculation of routing table is needed */
    boolean reroute=true;
    /** a discrete-event simulator */
    SimEventProcessor seq;
    /** The Physical Layer Simulator */
	public PhysicalLayer phyLayer;
	public PhysicalLayer wl_phyLayer;
	
	int multicast_hop_limit = 255;
	
	public void setMulticastHopLimit(int limit){
		multicast_hop_limit = limit;
	}
	
	/** Nodes that are currently selected by mouse */
	protected Hashtable selectedNodes = new Hashtable();
	
	public void setSelectedNodes(XYAddress start, XYAddress end, boolean append){
		if(!append){
			selectedNodes.clear();
		}
		for(int i=0;i<nodes.size();i++){
			OverlayNode onode = (OverlayNode)nodes.get(i);
			XYAddress addr = onode.getAddress();
			if(addr.x>=start.x && addr.y>=start.y && addr.x<=end.x && addr.y<=end.y){
				if(isNodeSelected(onode)){
					selectedNodes.remove(onode);
				}else{
					selectedNodes.put(onode, onode);
				}
			}
		}
	}
	
	public void setSelectedNode(OverlayNode onode, boolean append) {
		if(!append){
			selectedNodes.clear();
		}
		if(onode!=null){
			if(isNodeSelected(onode)){
				selectedNodes.remove(onode);
			}else{
				selectedNodes.put(onode, onode);
			}
		}
	}
	public boolean isNodeSelected(OverlayNode onode){
		return selectedNodes.containsKey(onode);
	}
	public OverlayNode getFirstSelectedNode(){
		Enumeration e = selectedNodes.elements();
		if(e.hasMoreElements()){
			return (OverlayNode)e.nextElement();
		}
		return null;
	}
	OverlayNode [] getSelectedNodes(){
		OverlayNode [] allSelected = new OverlayNode[selectedNodes.size()];
		Enumeration e = selectedNodes.elements();
		int i=0;
		while(e.hasMoreElements()){
			allSelected[i] = (OverlayNode)e.nextElement();
			i++;
		}
		return allSelected;
	}
	
	int getNumOfLeaf(){
		int num=0;
		if(active_nodes!=null){
			for(int i=0;i<active_nodes.size();i++){
				OverlayNode onode=(OverlayNode)active_nodes.get(i);
				if(onode.logical_neighbors.size()<=1)num++;
			}
		}	
		return num;
	}
	
	/** Forwarding message? */
	boolean forwarding=true;
	 
    /**Constructor
     * @param ta    a physical network topolog
     */
    OverlayNetwork( TopoAdaptor ta ){
    	//setSimTime(0f);
		
        phyLayer = new SimplePhysicalLayer(this);
        wl_phyLayer = new WirelessPhysicalLayer(this);
        
        
        nodes = new Vector();
        active_nodes = new Vector();
        seq = new SimEventProcessor();

        int i;
        PhyNode[] pnodes = ta.getNodes();

        for(i=0; i<pnodes.length;i++){
            addNode( pnodes[i]);
        }
            
        /**
        for(i=0; i<nodes.size();i++){
            fireMenu((OverlayNode)nodes.elementAt(i),0);
        }
        */
        
        PhyEdge[] pedges = ta.getEdges();
        if( null != pedges ) {
            for(i=0; i<pedges.length; i++){
                //lotusDebut.println("add edge:" + pedges[i].GetSrc() + ":" + pedges[i].GetDst());
                addEdge( (OverlayNode) nodes.elementAt(pedges[i].GetSrc()),
                     (OverlayNode) nodes.elementAt(pedges[i].GetDst()),
                     pedges[i].GetDistance() );
            }
        }
        
		
    }

    /**
     * Constructor
     */
    OverlayNetwork( ) {
		//setSimTime(0.0f);
		
        phyLayer = new SimplePhysicalLayer(this);
        wl_phyLayer = new WirelessPhysicalLayer(this);
        
        nodes = new Vector();
        active_nodes = new Vector();
        seq = new SimEventProcessor();
    }
	
	/**
     * Build the physical layer routing table
     */
    public void buildPhysicalRoute(){
    	
		updatePhysicalAdjacency();
    	
        if( reroute == false ) return;
        reroute = false;
        // This function clears the unicast routing
        // tables and then calls rebuildNodeRoute
        // on each of the nodes to always build a
        // shortest path route.

        // clear away the old tables
        //lotusDebug.errorln("rebuilding routing table");
        
        try{
        	for (int i = 0; i < nodes.size(); i++)
            	((OverlayNode) nodes.elementAt(i)).clearRoutingTable();

        	for (int i = 0; i < nodes.size(); i++)
        	{
	        	float diameter = rebuildNodeRoute((OverlayNode) nodes.elementAt(i));
        	//if( i%200 == 0 ) lotusDebug.println("round:"+i);
	        /*if( i % 200 == 0 ) {
	            Runtime.getRuntime().gc();
	            lotusDebut.println("free memory:" + Runtime.getRuntime().freeMemory() );
	        }*/
	
		    }
		}catch(Exception e)
		{
			lotusDebug.errorln("Buidling physical route: "+e.toString());
		}

    }
	/**
     * build the routing table entries for a specific node
     */
    final float rebuildNodeRoute(OverlayNode dst)
    {
        // This function rebuilds the route from every node
        // to the destination node.
        // Its a slower variant of Dijkstra's algorithm.
        // It assumes all edges are bidirectional.

        L_PriorityQueue_Dijkstra edges = new L_PriorityQueue_Dijkstra(nodes.size());
        Hashtable hop_table = new Hashtable();


	//System.out.print("Routing To Node: ");
	//lotusDebut.println(dst.getAddress().toString());

        // insert nodes which are neighbors of the destination node.
        Enumeration neighbors = dst.getPhyNeighbor();
        int round=0;
        while (neighbors.hasMoreElements())
        {

            OverlayNode n = (OverlayNode) neighbors.nextElement();
            edges.insert(dst.getDistanceTo(n), n);
            hop_table.put(n, dst);
        }

        float diameter = (float) 0.0;

        while (!edges.isEmpty())
        {
            float distance_from_dst = edges.minKey();
            OverlayNode src = (OverlayNode) edges.removeMin();

            // if src does not already have a routing entry to the destination node.
            if (src.getPhyNextHop(dst) == null)
            {
                src.insertIntoRoutingTable(dst,(OverlayNode)hop_table.get(src), distance_from_dst);
        		diameter = distance_from_dst;

                if (false)
                {
                    System.out.print("From ");
                    System.out.print(src.getAddress().toString());
                    System.out.print(" to ");
                    System.out.print(dst.getAddress().toString());
                    System.out.print(" via ");
                    System.out.println(((OverlayNode)hop_table.get(src)).getAddress().toString());
                    System.out.print("; Sum distance =  ");
                    System.out.println(distance_from_dst);
                }

                // insert edges out of src
                if(!forwarding)continue; //don't need to calculate more than one hop
                
                neighbors = src.getPhyNeighbor();
                while (neighbors.hasMoreElements())
                {
                    OverlayNode n = (OverlayNode) neighbors.nextElement();

                    if (n == dst)
                        continue;

                    float distance = distance_from_dst + src.getDistanceTo(n);
                    if (edges.getKeyOf(n) < 0)
                    {
                        edges.insert(distance, n);
                        hop_table.put(n, src);
                    }
                    else if (distance < edges.getKeyOf(n))
                    {
                        edges.setKeyOf(n, distance);
                        hop_table.remove(n);
                        hop_table.put(n, src);
                    }
                }

            }
        }
        /*edges = null;
        hop_table = null;*/
        return diameter;
    }
    /**
     * add an physical edge. Note, edges are always bi-directional
     * @param from starting node
     * @param to   ending node
     * @param distance the lenght of the edge
     */
    public boolean addEdge( OverlayNode from, OverlayNode to, float distance ){
	//add edges in a symmetrically way, i.e. always add two edges
        if( from.hasPhyEdge(to) ) return false;
		if( !from.wired_link || !to.wired_link)return false;
		
        reroute = true;
        from.insertPhyEdge( to, distance );
        to.insertPhyEdge( from, distance );
        return true;
    }

    /**
     * @return the vector containing the nodes in the overlay network
     */
    public Vector getActiveNodes() { return active_nodes;}

    /**
     * @return the vector containing all the nodes
     */
    public Vector getPhyNodes() { return nodes;}

    /*public boolean addLogicalEdge( OverlayNode from, OverlayNode to ){
	//add edges in a symmetrically way, i.e. always add two edges
        if( from.hasLogicalEdge(to) ) return false;

        from.insertLogicalEdge( to );
        to.insertLogicalEdge( from );
        return true;
    }*/

    /**
     * add a physical node with a specified address
     * @param addr the physical address of the node
     */
    public OverlayNode addNode( XYAddress addr ){
        reroute = true;
	    Enumeration e = nodes.elements();
    	OverlayNode onode;
	    boolean unique = true;

    	while( e.hasMoreElements() ){
            onode = (OverlayNode)e.nextElement();
            if( onode.getAddress().equals(addr) ){
            lotusDebug.errorln(" non unique node insertion detected");
            lotusDebug.errorln(" non unique address:" + addr.getX() + ":" + addr.getY() );
            unique = false;
            break;
	        }
    	}

        onode = new OverlayNode( addr, this,  null);
        onode.forwarding=forwarding;
        
        nodes.addElement( onode );
        return onode;

    }
	
	  /**
	   * added in Oct 25 for mobile node
     * add a physical node with a specified address
     * @param addr the physical address of the node
     */
    public OverlayNode addNode( PhyNode pnode ){
    	    XYAddress addr = new XYAddress(pnode.GetX(),pnode.GetY());									   
	    Enumeration e = nodes.elements();
    	OverlayNode onode;
	    boolean unique = true;
		

    	while( e.hasMoreElements() ){
            onode = (OverlayNode)e.nextElement();
            if( onode.getAddress().equals(addr) ){
            lotusDebug.errorln(" non unique node insertion detected");
            lotusDebug.errorln(" non unique address:" + addr.getX() + ":" + addr.getY() );
            unique = false;
            break;
	        }
    	}

        onode = new OverlayNode( pnode, this, null);
        onode.forwarding=forwarding;
        
        nodes.addElement( onode );
        return onode;

    }
    /**
     * remove a node in the network
     * @param node the node to be removed
     * @reutrn ture if the node exists, false otherwise
     */
    public boolean removeNode( OverlayNode node ){
        if( nodes.contains(node) == false ) return false;
        reroute = true;
        Enumeration enumTmp = node.getPhyNeighbor();
        while( enumTmp.hasMoreElements() ) {
            OverlayNode onode = (OverlayNode) enumTmp.nextElement();
            onode.removePhyEdge(node);
        }
        if(node.getProtocol()!=null)
        	node.getProtocol().resetProtocol();
        nodes.remove(node);
        active_nodes.remove(node);
	    return true;
    }

    /**
     * remove an edge
     * @param from the starting node
     * @param to the ending node
     */
    public boolean removeEdge( OverlayNode from, OverlayNode to ){
        if( from.hasPhyEdge(to) == false || to.hasPhyEdge(from) == false ) return false;
        from.removePhyEdge(to);
        to.removePhyEdge(from);
        reroute = true;
	    return true;
    }

    /**
     * An abstract function, which must be implemented by each overlay
     * protocol.
     * @param src the source node
     * @param dst the destination node
     * @return the next hope node in the logical network
     */
    public abstract OverlayNode getLogicalNextHop(OverlayNode src, OverlayNode dst);
	
	
	
    /**
     * find the closest node to a physical location. Used in GUI interface
     */
    public OverlayNode findClosest( XYAddress addr, int radius ){
        int i;
        float dist=Float.MAX_VALUE;
        int index=-1;

        for( i=0; i<nodes.size(); i++){
            if( dist > addr.distanceTo( ((OverlayNode)nodes.elementAt(i)).getAddress()  ) ){
            dist = addr.distanceTo( ((OverlayNode)nodes.elementAt(i)).getAddress()  );
            index=i;
            }
        }

        if( -1 == index ) return null;
        if( dist > radius ) return null;

        return (OverlayNode)nodes.elementAt(index);

    }

    /**
     * find the edge which is the closest to a physical address
     */

    public boolean findClosestLine( XYAddress addr, int radius, OverlayNode []nodeArray ){
        double dista;
        double distb;
        double distc;
        double cos1;
        double cos2;
        double sin1;

        for( int i=0; i<nodes.size(); i++){
            for( int j=0; j<i; j++){
                OverlayNode fromnode = (OverlayNode) nodes.elementAt(i);
                OverlayNode tonode = (OverlayNode) nodes.elementAt(j);

                dista = addr.distanceTo(fromnode.getAddress());
                distb = addr.distanceTo(tonode.getAddress());
                distc = fromnode.getAddress().distanceTo(tonode.getAddress());

                /* the structure

                        from ------------c-------------to
                           - 1                       2-
                             -                      -
                               a                 b
                                 -            -
                                    -      -
                                       node

                    */
                cos1 = (dista*dista + distc*distc - distb*distb)/(2*dista*distc);
                cos2 = (distb*distb + distc*distc - dista*dista)/(2*distb*distc);

                if( cos1 < 0 || cos2 < 0 ) continue;
                sin1 = Math.sqrt(1.0 - cos1*cos1);
                if( dista*sin1 <= radius ) {
                    if( fromnode.hasPhyEdge(tonode) ) {
                        nodeArray[0] = fromnode;
                        nodeArray[1] = tonode;
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * @return the physical topology of the network
     */
    public TopoAdaptor returnTopo( ){
	    TopoAdaptor_HM topo = new TopoAdaptor_HM();

	    topo.nodes = new PhyNode[nodes.size()];

        int numedges=0;
        Enumeration e = nodes.elements();

        while( e.hasMoreElements() ){
            OverlayNode onode =(OverlayNode) e.nextElement();
            numedges += onode.getWiredNeighbors().size();
        }

        topo.edges = new PhyEdge[numedges];

        int i, j;
        for( i=0,j=0; i<nodes.size(); i++){

            OverlayNode fromnode = (OverlayNode) nodes.elementAt(i);
            
            topo.nodes[i] = new PhyNode(
                        fromnode.getAddress().getX(),
                        fromnode.getAddress().getY(),
                        fromnode.getMobilityStr(),
                        fromnode.mobile_speed,
                        fromnode.isWired(),
                        fromnode.isWireless() );

            /*if( active_nodes.contains( fromnode )){
            topo.activeNodes.addElement( topo.nodes[i] );
            }*/

            Enumeration neighbor = fromnode.getWiredNeighbors().elements();

            while( neighbor.hasMoreElements() ){
            OverlayNode tonode = (OverlayNode) neighbor.nextElement();

            topo.edges[j++] = new PhyEdge(
                            i,
                            nodes.indexOf(tonode),
                            fromnode.getDistanceTo(tonode) );

            }
        }

        return topo;
    }

	/**
	 * Update the Adjacency list of node to node connectivity
	 */
	public void updatePhysicalAdjacency()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((OverlayNode) nodes.elementAt(i)).updatePhysicalNeighbors();
	}
	public void updatePhysicalNodes()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((OverlayNode) nodes.elementAt(i)).move();
	}
	public void MACStep()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((OverlayNode) nodes.elementAt(i)).MAC.MACStep();
	}

    /**
     * @return the simulation time
     */
    final float simTime(){ return seq.getSimTime(); }
    final float getSimTime(){ return simTime(); }
    final long getMSSimTime(){ return seq.getMSSimTime(); }

    /**
     * insert an simulation event to the discrete event simulator
     */
    final void insertEvent( SimEvent e ) { seq.insertEvent(e); }
	final void clearEvent(SimEvent e) {seq.clearEvent(e);};
	final float getEvent(SimEvent e) {return seq.getEvent(e);}
	
	
    /**
     * get all the pending events
     */
    final Enumeration getPendingEvent( ) { return seq.getPendingEvents(); }

    /**
     * advance the simulation to the next pending event
     */
	final void step()
	{
		if(seq.queueIsEmpty())return;
		
		SimEvent se;
		float sim_time = getSimTime();
		float end_time = sim_time + SIM_STEP;

		while(end_time > sim_time)
		{	
			this.buildPhysicalRoute();
			this.updatePhysicalNodes();
			this.updatePhysicalAdjacency();
			this.MACStep();
			
			if(seq.queueIsEmpty())return;
			
			while( (se=seq.getNextEvent(end_time)) != null )
			{
				
				if( (se == null) )
				{
					
					//lotusDebug.errorln("find a null event");
					
				}
				else
				{
					//sim_time=se.getTime();
					sim_time=se.getTime();
					
					if(se.getRecipient()==null)
					{
						//se.getSender().receiveEvent(se);
						for(int i =0;i <se.getSender().getWirelessNeighbors().size();i++)
						{
							OverlayNode temp = (OverlayNode)se.getSender().getWirelessNeighbors().elementAt(i);
							if(temp!=se.getSender())
								temp.receiveEvent(se);
						}
					}
					else if(se.getRecipient().isNeighbor(se.getSender())){
						se.getRecipient().receiveEvent(se);
					}
					
					if((se.eventType() == se.MESSAGE_EVENT)){
						//lotusDebug.errorln(""+sim_time+":"+se);
						return;
					}
				}
			}
			
			//setSimTime(end_time);
			sim_time = end_time;
			end_time = sim_time + SIM_STEP;
			
		}
	}

    /**
     * run the simulation for the specific amount of time
     * @param time the duration to run the simulation
     */
   final void advance(float adv_time)
	{
		float sim_time=getSimTime();
		float big_end_time=sim_time+adv_time;
		while(sim_time<big_end_time){
			SimEvent se;
			float end_time = sim_time + SIM_STEP;
			
			this.buildPhysicalRoute();
			this.updatePhysicalNodes();
			this.updatePhysicalAdjacency();
			this.MACStep();
			
			while( (se=seq.getNextEvent(end_time)) != null )
			{
				//setMSSimTime(se.getMSTime());
				sim_time=se.getTime();
				
				if(se.getRecipient()==null)
				{
					//se.getSender().receiveEvent(se);
					for(int i =0;i <se.getSender().getWirelessNeighbors().size();i++)
					{
						OverlayNode temp = (OverlayNode)se.getSender().getWirelessNeighbors().elementAt(i);
						
						//lotusDebug.errorln("sent to one recipient");
						
						if(temp!=se.getSender())
							temp.receiveEvent(se);
					}
				}
				else if((se.getRecipient().isNeighbor(se.getSender())) )
				{
					if((se.eventType() == se.MESSAGE_EVENT)){
						//lotusDebug.errorln(""+sim_time+":"+se+" sent to "+se.getRecipient());
						
					}
					
					se.getRecipient().receiveEvent(se);
				}
				//sim_time=se.time;
				//lotusDebug.errorln(""+sim_time+":"+se);
			}
			//setSimTime(end_time);
			sim_time = end_time;
		}
		//lotusDebug.errorln("advance finished");
		
	}
    //start the simulation: start the timers of each node and sending the Iam leader to
    //the server
    boolean simRunning = false;

    /**
     * Abstract function, used by each overlay protocol to implement those
     * initialze functions before the staring of the simulation
     */
    abstract boolean start();
    /**
     * Abstract function, used by each overlay protocol to clean up the current simulation
     * and restore the simulator to its default state
     */
    void reset(){
    	this.seq.reset();
		
    	for(int i=0;i<this.nodes.size();i++){
    		OverlayNode onode = (OverlayNode)nodes.elementAt(i);
    		onode.MAC.MACInit();
    		onode.wl_MAC.MACInit();
    	}
    }
    /**
     * @return whether the simulation is running
     */
    boolean isRunning() { return simRunning; }

    /**
     * An abstract function, implemented by each overlay protocol to build
     * its own overaly topology. This function actually implements the
     * static simulation
     */
    abstract boolean buildOverlay();

    //the protocol-dependent part of the GUI interface part
    /**
     * Abstract function, used by each protocol to implement protoc-specific
     * functions
     */
    public abstract PopupMenu getMenus( );
    
   
    public static PopupMenu convertStringsToMenu(String[] menuName, ActionListener al){
    	PopupMenu pmenu = new PopupMenu();
    	MenuItem []mi_sim = new MenuItem[menuName.length];
        for( int i=0; i<menuName.length; i++){
        	mi_sim[i] = new MenuItem( menuName[i] );
            pmenu.add(mi_sim[i]);
            mi_sim[i].addActionListener(al);
        }
        return pmenu;
    }
    public static int getIndexOfMenu(String menu, String[] menu_names){
    	for(int i=0;i<menu_names.length;i++){
    		if(menu_names[i].equals(menu))
    			return i;
    	}
    	return -1;
    }
    abstract public boolean activateNode( OverlayNode o );
    abstract public boolean deactivateNode(OverlayNode o);
    
    public void setSpeed(float speed){
    	for(int i=0;i<nodes.size();i++){
    		OverlayNode onode = (OverlayNode)nodes.get(i);
    		onode.setMobileSpeed(speed);
    	}
    }
    /**
     * Abstract function, it is the callback function invoked by the simulator
     * and implemented by each overlay protocol to execute protocol-specific
     * operations
     */
    //public abstract boolean fireMenu( OverlayNode onode, int index );

    
	//the protocol menu and instance creation part
    final static String [] theMobilemode = {
        //"Random Motion",
		  "Linear Motion",
    };
    
    /** @return current supported protocols */
    static String[] getMobilemode(){
	    return theMobilemode;
    }
	//the protocol menu and instance creation part
    final static String [] theProtocols = {
		"Hypercast Protocols",
		/*
        "Ad Hoc SPT",
        "Ad Hoc DT-precise",
        "sample",
        "DT-precise",
        "DT-random",
        "DT-Geo",
        "DT-LM",
        "DT-Truc",
        "MST",
        "SPT",
        "Narada",
        */
        //"MAC Tester"
        
    };
    
    /** @return current supported protocols */
    static String[] getProtocols(){
	    return theProtocols;
    }
	
    /**
     * Create a simulation network
     */
    public static OverlayNetwork createInst(int index){
        switch (index) {
			case 0: return new simHCNetwork();
			/*
            case 1: return new simAHSPT();
          	case 2: return new simAHDTPrecise();
          	case 3: return new simTree();
          	case 4: return new simDTPrecise();
          	case 5: return new simDTRandom();
          	case 6: return new simDTGeo();
            case 7: return new simDTLandmark();
            case 8: return new simDTTruncation();
            case 9: return new simMST();
            case 10: return new simSPT();
            case 11: return new simNarada();
            */
            //case 12: return new simAHTester();
            
        }
        return null;
    }

    /**create a simulation network
     * @param index the type of simulation created
     * @param ta the physical topology of the network
     */
    static OverlayNetwork createInst(int index, TopoAdaptor ta){
        switch (index) {
			case 0: return new simHCNetwork(ta);
			/*
            case 1: return new simAHSPT(ta);
            case 2: return new simAHDTPrecise(ta);
            case 3:  return new simTree(ta);
            case 4: return new simDTPrecise(ta);
          	case 5: return new simDTRandom(ta);
          	case 6: return new simDTGeo(ta);
            case 7: return new simDTLandmark(ta);
            case 8: return new simDTTruncation(ta);
            case 9: return new simMST(ta);
            case 10: return new simSPT(ta);
            case 11: return new simNarada(ta);
            */
            //case 12: return new simAHTester(ta);
			
        }
        return null;
    }

	
	
    //import/export protocol-specific protocols
    /**
     * An abstract function, implemented by each protocol to read topology description file
     * with protocol-specific information
     */
    abstract boolean protocolReader( String fname );
    
    /**
     * An abstract function, implemented by each protocol to save topology description file
     * with protocol-specific information
     */
    abstract boolean protocolWriter( String fname );
    
}

