/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.awt.Color;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Overlay Node, composed of physical and logical parts
 * Main components </p>
 * a. physical address </p>
 * b. physical links </p>
 * c. logical links </p>
 * d. physical name </p>
 * e. routing table to provide the physical network service </p>
 * f. protocol(logical part) </p>
 * Main functions </p>
 * a. maintenence of the physical/logical links </p>
 *      note: the physical links are added/deleted by outside
 *            the logical links are added/deleted by the protocol/graph-sim
 * b. dispatch/receive the events for two layers. </p>
 *          interface to the event dispatcher: send/receive the event(timer/msg)
 *          interface to the protocol: send/receive the event(timer/msg)
 *          for the message, it may need to route the message to the next hop
 * c. add/delete/start/end the running protocol </p>
 */
/**
 * revised by Guangyu Dong, Jan 6, 2004
 * to add mobility and wireless support
 */

class OverlayNode implements physicalInterface {

	/** Start of addon by Guangyu to support mobility and wireless network*/
	/** Mobility Mode */
	MobilityInterface mobility;

	/** There's a wired link? */
	boolean wired_link = true;
	/** There's a wireless link? */
	boolean wireless_link = false;

	/** Default moving speed */
	static final float DEF_MOBILE_SPEED = 0.2f;

	/** wireless transmission range */
	float wireless_trans_radius = WirelessPhysicalLayer.SIGNAL_RANGE;
	/** wireless sensing range */
	float wireless_sense_radius = WirelessPhysicalLayer.SENSE_RANGE;
	/** moving speed of this node */
	float mobile_speed = DEF_MOBILE_SPEED;

	/** neighbors inside the wireless transmission range */
	Vector wireless_neighbors;
	/** distances to wireless neighbors */
	Vector wireless_distances;

	/** neighbors inside the wireless sensing range */
	Vector wireless_sensing_neighbors;

	/* Forwarding message??
	 * This flag decided whether this network is a multihop network
	 * If yes, every node will work as a rounter.
	 */
	boolean forwarding = true;

	/**
	 * @return whether the node has wireless link
	 */
	public boolean isWireless() {
		return wireless_link;
	}
	/**
	 * @return whether the node has wired link
	 */
	public boolean isWired() {
		return wired_link;
	}
	/**
	 * @return ths string representation of the mobility mode
	 */
	public String getMobilityStr() {
		if (mobility == null)
			return null;
		else
			return mobility.toString();
	}
	/*
	 * Set the mobility mode
	 */
	public void setMobility(String mobilityMode) {
		//lotusDebug.errorln("set mobility "+mobilityMode);
		setMobility(mobilityMode, DEF_MOBILE_SPEED);
		//lotusDebug.errorln("mobility= "+mobility);
	}
	/**
	 * Set the mobility mode and the speed
	 */
	void setMobility(String mobilityMode, float speed) {

		if (mobilityMode == null) {
			mobility = null;
			mobile_speed = 0;
			return;
		}

		//lotusDebug.errorln("mobile_speed="+mobile_speed);
		//lotusDebug.errorln("speed="+speed);
		mobile_speed = speed;

		if (mobilityMode.equals("Random Motion"))
			mobility = new MobiRandomStep(speed);
		else if (mobilityMode.equals("Linear Motion"))
			mobility = new MobiReflectiveLinear(speed);
		else {
			mobility = null;
		}
	}

	/*
	 * @return if this node is mobile
	 */
	public boolean isMobile() {
		return mobility != null;
	}
	/*
	 * set the speed
	 */
	public void setMobileSpeed(float sp) {
		mobile_speed = sp;
		if(this.mobility!=null)
			setMobility(this.mobility.toString(), sp);
	}
	/*
	 * set wireless transmission range
	 */
	public void setRadius(float tr) {
		wireless_trans_radius = tr;
	}
	/*
	 * @return wireless transmission range
	 */
	public float GetRadiusF() {
		return wireless_trans_radius;
	}
	/*
	 * @return integer value of wireless transmission range
	 */
	public int GetRadius() {
		return (int)wireless_trans_radius;
	}
	/*
	 * Set link types
	 * @param wd wired link flag
	 *        wl wireless link flag
	 */
	private void setLinkType(boolean wd, boolean wl) {
		wired_link = wd;
		wireless_link = wl;

		if (!wired_link) {
			//clear neighbors
			physical_neighbors = new Vector();
			physical_distances = new Vector();

		}
		if (!wireless_link) {
			//clear neighbors
			wireless_neighbors = new Vector();
			wireless_distances = new Vector();

			wireless_sensing_neighbors = new Vector();

		}
	}
	/*
	 * Set link type
	 * @param type 
	 */
	public void setLinkType(int type) {
		switch (type) {
			case INTERFACE_WIRED :
				setLinkType(true, false);
				break;
			case INTERFACE_WIRELESS :
				setLinkType(false, true);
				break;
			case INTERFACE_BOTH :
				setLinkType(true, true);
				break;
		}
	}

	/*
	 * Move the node for 1 second
	 */
	public void move() {
		if (!isMobile())
			return;

		if (mobility != null) {
			Position prev = new Position(addr.getXF(), addr.getYF(), 0);
			Position next = mobility.move(net.simTime(), prev);
			moveTo(next.getX(), next.getY());
		}
	}
	/*
	 * Move to a specific location
	 */
	public void moveTo(float x, float y) {
		//lotusDebug.errorln(""+this+"move to "+x+","+y);
		addr = new XYAddress(x, y);
	}
	/**
	 * @return geographic distance to another node
	 */
	public final float getDirectDistanceTo(OverlayNode n) {
		if (n == this)
			return (float)0.0;

		return (float)Math.sqrt(
			Math.pow(n.addr.x - addr.x, 2) + Math.pow(n.addr.y - addr.y, 2));
	}

	/**
	 * get the multi-hop netowrk distance to a node
	 * @param n the destination node
	 * @return the distance
	 */
	public final float getDistanceTo(OverlayNode n) {
		if (n == this)
			return (float)0.0;

		int index = physical_neighbors.indexOf(n);
		if (index != -1)
			return ((Float)physical_distances.elementAt(index)).floatValue();
		else if (isWirelessNeighbor(n))
			return getDirectDistanceTo(n);

		if (routing_nexthop == null)
			return -1.0f;
		index = net.getPhyNodes().indexOf(n);
		if (index != -1) {
			float dist = 0;
			OverlayNode nextnode = routing_nexthop[index];
			if (null == nextnode)
				return -1.0f;
			OverlayNode prevnode = this;
			while (nextnode != n) {
				dist += prevnode.getDistanceTo(nextnode);
				prevnode = nextnode;
				nextnode = nextnode.getPhyNextHop(n);
			}
			dist += prevnode.getDirectDistanceTo(nextnode);
			return dist;
		} else {
			lotusDebug.errorln("can not route between the two points");
			return -1.0f;
		}

		//return Float.MAX_VALUE;

	}

	/*
	 * @return if otherNode is a wired network neighbor 
	 */
	boolean isWiredNeighbor(OverlayNode otherNode) {
		if (!wired_link)
			return false;
		if (!otherNode.wired_link)
			return false;

		return physical_neighbors.indexOf(otherNode) >= 0;
	}
	/*
	 * @return if otherNode is inside the sensing range
	 */
	boolean isSensingNeighbor(OverlayNode otherNode) {
		if (!wireless_link)
			return false;
		if (!otherNode.wireless_link)
			return false;

		float distance = getDirectDistanceTo(otherNode);
		if ((distance <= wireless_sense_radius))
			return true;
		else
			return false;
	}
	/*
	 * @return if otherNode is inside the wireless transmission range
	 */
	boolean isWirelessNeighbor(OverlayNode otherNode) {
		if (!wireless_link)
			return false;
		if (!otherNode.wireless_link)
			return false;

		float distance = getDirectDistanceTo(otherNode);
		if ((distance <= wireless_trans_radius))
			return true;
		else
			return false;
	}
	/*
	 * @return if otherNode is a network neighbor
	 */
	boolean isNeighbor(OverlayNode otherNode) {

		if (otherNode == this)
			return true;

		return isWiredNeighbor(otherNode) || isWirelessNeighbor(otherNode);

	}

	/**
	 * insert a wireless edge for this node
	 * @param n the new neigbhor
	 * @param weight the length of the node
	 */
	boolean insertWirelessEdge(OverlayNode n, float weight) {
		if (!wireless_link)
			return false;

		if (wireless_neighbors.contains(n))
			return false;
		else {
			wireless_neighbors.addElement(n);
			wireless_distances.addElement(new Float(weight));
		}
		return true;
	}
	/**
	 * remove an wireless edge
	 * @param n the node at the other end of the edge
	 * @return true if such a neighbor exists, false otherwise
	 */
	boolean removeWirelessEdge(OverlayNode n) {
		if (wireless_neighbors.contains(n)) {
			int index = wireless_neighbors.indexOf(n);
			wireless_neighbors.remove(index);
			wireless_distances.remove(index);
			return true;
		} else {
			return false;
		}
	}

	/*
	 * @return all wired neighbors
	 */
	public Vector getWiredNeighbors() {
		return physical_neighbors;
	}
	/*
	 * @return physical edge length to otherNode
	 */
	public float getPhyEdgeLen(OverlayNode otherNode) {
		int index = getAllNeighbors().indexOf(otherNode);
		if (index >= 0)
			return getPhyEdgeLen(index);
		else
			return -1;
	}
	/*
	 * @return wired edge length of some neighbor
	 */
	public float getWiredEdgeLen(int index) {
		return ((Float)physical_distances.elementAt(index)).floatValue();
	}
	/*
	 * @return all wireless neighbors
	 */
	public Vector getWirelessNeighbors() {
		return wireless_neighbors;
	}
	/*
	 * @return all nodes inside the wireless sensing range 
	 */
	public Vector getSensingNeighbors() {
		return wireless_sensing_neighbors;
	}

	/** @return all neighbors, including wired and wireless neighbors */
	public Vector getAllNeighbors() {
		Vector all = new Vector();
		if (wired_link)
			all.addAll(physical_neighbors);
		if (wireless_link)
			all.addAll(wireless_neighbors);

		return all;
	}
	/** @return all distances */
	public Vector getAllDistances() {
		Vector all = new Vector();
		if (wired_link)
			all.addAll(physical_distances);
		if (wireless_link)
			all.addAll(wireless_distances);

		return all;
	}
	/**
	 * Update the wireless neighbor list
	 */
	public void updatePhysicalNeighbors() {

		if (!wireless_link)
			return;

		/*
		float distance;
		
		Enumeration enumTmp = wireless_neighbors.elements();
		while( enumTmp.hasMoreElements() ) {
		   otherNode = (OverlayNode) enumTmp.nextElement();
		   removeWirelessEdge(otherNode);
		}
		*/
		OverlayNode otherNode;

		//clear the wireless neighbor
		wireless_neighbors.clear();
		wireless_distances.clear();
		wireless_sensing_neighbors.clear();

		for (int i = 0; i < net.nodes.size(); i++) {
			otherNode = (OverlayNode)net.nodes.elementAt(i);
			if (isWirelessNeighbor(otherNode))
				insertWirelessEdge(otherNode, getDirectDistanceTo(otherNode));
			else if (isSensingNeighbor(otherNode))
				wireless_sensing_neighbors.addElement(otherNode);
		}
	}

	/** predefined link type */
	static final int INTERFACE_WIRED = 0;
	static final int INTERFACE_WIRELESS = 1;
	static final int INTERFACE_BOTH = 2;

	/** End of addon by Guangyu */

	/** wired link packet speed (m/s) */
	final static float PACKET_SPEED = 1000.0f;

	/** node attributes */
	/** name of the node */
	String name;

	/** physical node*/
	//disable by Guangyu
	//because it's not good to use it here
	//PhyNode phyNode;

	/** physical address of the node
	 *  it's actually the location on the GUI 
	 */
	XYAddress addr;

	//logical structure
	/**
	 * protocol running on this node. It can be empty
	 * if no protocol is running
	 */
	protocolInterface protocol;
	/** neighbors in the overlay */
	Vector logical_neighbors; // logical neighbors
	/** color used to display edges in the overlay */
	Vector logical_colors; // the link color, display purpose
	/** the width of the logical edges */
	Vector logical_width; // the link width, display purpose;

	//physical structure
	/** physical neighbors */
	Vector physical_neighbors; // physical neighbors
	/** the lenght of the physical edges */
	Vector physical_distances; // physical neighbors

	//routing table, the next hop is indexed by the node's position
	//in the OverlayNetwork's nodes vector
	/** routing table at this node. It stores the next hope node from this
	 *  node to all the other nodes
	 *
	 */
	OverlayNode[] routing_nexthop;
	/** a referece to the class Overlay Network */
	OverlayNetwork net;

	/** MAC Layer simulator */
	public MACInterface MAC;
	public MACInterface wl_MAC;

	/** time when the last message sent by this node will be delivered (junk data) */
	public float timeLastMsg;

	/** last time this node stop sending or receving data */
	//public float wl_timeLastSent=-10f;
	//public float wl_timeLastReceived=-10f;
	public boolean wl_lastMsgCollision = false;

	/**constructor
	 * @param a node's physical address
	 * @param onet a reference to the overlay network
	 * @param sname node's name
	 */
	public OverlayNode(XYAddress a, OverlayNetwork onet, String sname) {
		//phyNode = new PhyNode(a.getX(), a.getY());

		addr = a;
		net = onet;
		name = sname;
		logical_neighbors = new Vector();
		logical_colors = new Vector();
		logical_width = new Vector();
		physical_neighbors = new Vector();
		physical_distances = new Vector();
		timeLastMsg = net.simTime();
		MAC = new SimpleMAC(this);
		wl_MAC = new AdhocMAC(this);

		wireless_neighbors = new Vector();
		wireless_distances = new Vector();
		wireless_sensing_neighbors = new Vector();
	}

	/**constructor
	 * @param a node's physical address
	 * @param onet a reference to the overlay network
	 * @param sname node's name
	 */
	public OverlayNode(PhyNode pnode, OverlayNetwork onet, String sname) {
		//phyNode = pnode;
		addr = new XYAddress(pnode.GetX(), pnode.GetY());
		net = onet;
		name = sname;
		logical_neighbors = new Vector();
		logical_colors = new Vector();
		logical_width = new Vector();
		physical_neighbors = new Vector();
		physical_distances = new Vector();
		timeLastMsg = net.simTime();
		MAC = new SimpleMAC(this);
		wl_MAC = new AdhocMAC(this);

		wireless_neighbors = new Vector();
		wireless_distances = new Vector();
		wireless_sensing_neighbors = new Vector();

		setMobility(pnode.mobility, pnode.speed);
		setLinkType(pnode.wired, pnode.wireless);

	}
	//attributes method
	/** @return node name */
	public String getName() {
		return name;
	}
	/** set the name of this node */
	public void setName(String sname) {
		name = sname;
	}
	/** @return the physical address of the node */
	public XYAddress getAddress() {
		//return new XYAddress(phyNode.GetX(),phyNode.GetY()); 
		return addr;
	}
	/** set the address of the node */
	public void setAddress(XYAddress a) {
		//phyNode.SetXF(a.getX());
		//phyNode.SetYF(a.getY());
		addr = a;
	}

	//logical & physical links' related method
	/**
	 * test if the node is its physical neighbor
	 */
	boolean hasPhyEdge(OverlayNode n) {
		if (physical_neighbors.contains(n))
			return true;
		else
			return false;
	}
	boolean hasPhyNeighbor(OverlayNode n) {
		if (hasPhyEdge(n))
			return true;
		if (wireless_neighbors.contains(n))
			return true;
		return false;
	}

	/**
	 * test if the node is its logical neighbor
	 */
	boolean hasLogicalEdge(OverlayNode n) {
		if (logical_neighbors.contains(n))
			return true;
		else
			return false;
	}

	/**
	 * insert a physical edge for this node
	 * @param n the new neigbhor
	 * @param weight the length of the node
	 */
	boolean insertPhyEdge(OverlayNode n, float weight) {
		if (!wired_link)
			return false;

		if (physical_neighbors.contains(n))
			return false;
		else {
			physical_neighbors.addElement(n);
			physical_distances.addElement(new Float(weight));
		}
		return true;
	}

	/**
	 * @return all the physical neighbors of this node
	 */
	Enumeration getPhyNeighbor() {
		//return physical_neighbors.elements();
		//return all neighbors instead
		return getAllNeighbors().elements();
	}

	/**
	 * @return all the logical neighbors of this node
	 */
	Enumeration getLogicalNeighbor() {
		return logical_neighbors.elements();
	}

	/**
	 * @return the colors of the logical links
	 */
	Enumeration getLogicalColor() {
		return logical_colors.elements();
	}

	/**
	 * @return the lengths of the logical links
	 */
	Enumeration getLogicalWidth() {
		return logical_width.elements();
	}

	/**
	 * @return the total number of logical neighbors
	 */
	int getLogicalNeighborNumber() {
		return logical_neighbors.size();
	}
	/**
	 * @return the total number of physical neighbors
	 */
	int getPhyNeighborNumber() {
		//return physical_neighbors.size(); 
		//reutrn getAllNeighbors() count instead
		return getAllNeighbors().size();
	}
	/**
	 * @return the index of this physical node in all the neighbors
	 */
	int getPhyNeighborIndex(OverlayNode onode) {
		//return physical_neighbors.indexOf( onode );
		//reutrn getAllNeighbors() instead
		return getAllNeighbors().indexOf(onode);
	}

	/**
	 * @return the legth of a physical edge, which is represented by
	 * its index among all the neighbors
	 */
	float getPhyEdgeLen(int index) {
		return ((Float)getAllDistances().elementAt(index)).floatValue();
	}

	/**
	 * remove an physical edge
	 * @param n the node at the other end of the edge
	 * @return true if such a neighbor exists, false otherwise
	 */
	boolean removePhyEdge(OverlayNode n) {
		if (physical_neighbors.contains(n)) {
			int index = physical_neighbors.indexOf(n);
			physical_neighbors.remove(index);
			physical_distances.remove(index);
			return true;
		} else {
			return false;
		}
	}

	//routing related functions

	/**
	 * clear the routing table of this node
	 */
	public void clearRoutingTable() {
		routing_nexthop = null;
		routing_nexthop = new OverlayNode[net.getPhyNodes().size()];
	}

	/**
	 * insert an routing table entry
	 */
	public void insertIntoRoutingTable(
		OverlayNode destination,
		OverlayNode next_hop,
		float distance) {
		int index = net.getPhyNodes().indexOf(destination);
		routing_nexthop[index] = next_hop;
	}

	/**
	 * @return the next hop node from this node to destination
	 */
	public OverlayNode getPhyNextHop(OverlayNode destination) {
		int index = net.getPhyNodes().indexOf(destination);
		//lotusDebut.println("inidex at: " + index );
		if (index < 0)
			return null;
		if (routing_nexthop == null)
			return null;
		if (routing_nexthop[index] == null)
			return null;
		return routing_nexthop[index];
	}

	/**
	 * clear all the logical edges
	 */
	public void clearLogicalEdges() {
		this.logical_colors.clear();
		this.logical_neighbors.clear();
		this.logical_width.clear();
	}

	//physicalInterface
	/**
	 * part of the physicalInterface implemenation
	 * @return the logical neighbors of this node
	 */
	public Enumeration logicalLinks() {
		return getLogicalNeighbor();
	}

	/**
	 * part of the physicalIniterface implementation
	 * @return the physical neighbors of this node
	 */
	public Enumeration physicalLinks() {
		return getPhyNeighbor();
	}

	/** <0 if the physical node is not a neighbor
	 *  otherwise, the link weight
	 */
	//float physicalWeight( physicalInterface );

	/**
	 * @the physical address of this node
	 */
	public XYAddress physicalAddress() {
		return getAddress();
	}

	/**
	 * insert a logical edge at this node
	 * @param n the other end of the logical edge
	 * @param color the color of the edge, for display purpose
	 * @param width the width of the edge, for dispaly purpose
	 */
	public boolean insertLogicalEdge(OverlayNode n, Color color, int width) {
		if (logical_neighbors.contains(n))
			return false;
		else {
			logical_neighbors.addElement(n);
			logical_colors.addElement(color);
			logical_width.addElement(new Integer(width));
		}
		return true;
	}

	/**
	 * remove a logical neighbor
	 */
	public boolean removeLogicalEdge(OverlayNode n) {
		if (logical_neighbors.contains(n)) {
			int index = logical_neighbors.indexOf(n);
			logical_neighbors.remove(index);
			logical_colors.remove(index);
			logical_width.remove(index);
			return true;
		} else {
			return false;
		}
	}

	//both the timer and the messages
	/**
	 * send a message, used by the logical part of the node
	 */
	public boolean sendMsg(SimEventMsg msg) {

		boolean sent = false;

		//lotusDebug.errorln(this.toString()+"going to send: "+msg+" to "+msg.getDst());

		OverlayNode dst = msg.getDst();
		//multicast message
		if (dst == null) {
			//Disable broadcast if hop limit is less than 1
			if(this.net.multicast_hop_limit<=0){
				return true;
			}
			msg.setTTL(this.net.multicast_hop_limit);
			if (wired_link) {
				//broadcast wired message
				//lotusDebug.errorln("src neighbor size="+physical_neighbors.size());
				
				for (int i = 0; i < physical_neighbors.size(); i++) {
					SimEventMsg wmsg = (SimEventMsg)msg.clone();
					wmsg.setTransType(SimEventMsg.TRANS_WIRED);
					wmsg.setSender(this);
					wmsg.setSrc(this);
					wmsg.setTTL(msg.getTTL());

					OverlayNode wdst = (OverlayNode)physical_neighbors.get(i);
					//float wdist = getWiredEdgeLen(i);
					float wdist = this.getDirectDistanceTo(wdst);
					
					//wmsg.setDst(wdst);
					wmsg.setRecipient(wdst);

					if (wmsg.eventType() == SimEvent.MESSAGE_EVENT) {
						//net.phyLayer.sendMsg(msg);
						//lotusDebug.errorln("sent: "+msg);
						wmsg.setTime(net.simTime() + wdist / PACKET_SPEED);
						net.insertEvent(wmsg);
						//MAC.sendMsg(wmsg);
						sent = true;
						//lotusDebug.errorln("sent wired broadcasting to "+wdst+":"+wmsg);
					} else {
						wmsg.setTime(net.simTime() + wdist / PACKET_SPEED);
						net.insertEvent(wmsg);
						sent = true;
					}
				}

			}

			if (wireless_link) {
				//broadcast wireless message
				msg.setTransType(SimEventMsg.TRANS_WIRELESS);

				msg.setSender(this);
				msg.setSrc(this);
				msg.setRecipient(null);
				if (msg.eventType() == SimEvent.MESSAGE_EVENT) {
					//net.phyLayer.sendMsg(msg);
					//lotusDebug.errorln("sent: "+msg);
					msg.setTime(
						net.simTime()
							+ wireless_trans_radius
								/ WirelessPhysicalLayer.PACKET_SPEED);
					//net.insertEvent( msg );
					wl_MAC.sendMsg(msg);
					sent = true;
				} else {
					msg.setTime(
						net.simTime()
							+ wireless_trans_radius
								/ WirelessPhysicalLayer.PACKET_SPEED);
					//net.insertEvent( msg );
					wl_MAC.sendMsg(msg);
					sent = true;
				}
			}
		}
		//unicast message
		else {
			OverlayNode nextHop = this.getPhyNextHop(dst);

			//if(!forwarding && nextHop!=dst)return;

			if (null == nextHop) {
				lotusDebug.errorln("packet can not be routed");
				return false;
			}
			float dist;

			boolean wired_msg = this.isWiredNeighbor(nextHop);

			if (wired_msg) {
				msg.setTransType(SimEventMsg.TRANS_WIRED);
				//dist = getPhyEdgeLen(nextHop);
				dist = this.getDirectDistanceTo(nextHop);
			} else {
				msg.setTransType(SimEventMsg.TRANS_WIRELESS);
				dist = wireless_trans_radius;
			}

			//lotusDebug.errorln(this.toString()+" is sending msg to "+nextHop);
			//lotusDebug.errorln("Message: "+msg+" to "+msg.getDst());

			msg.setSender(this);
			msg.setSrc(this);
			msg.setRecipient(nextHop);
			if (msg.eventType() == SimEvent.MESSAGE_EVENT) {
				//net.phyLayer.sendMsg(msg);

				if (wired_msg) {
					msg.setTime(net.simTime() + dist / PACKET_SPEED);
					net.insertEvent(msg);
				} else {
					msg.setTime(
						net.simTime()
							+ dist / WirelessPhysicalLayer.PACKET_SPEED);
					wl_MAC.sendMsg(msg);
				}
				sent = true;
			} else {
				//???
				msg.setTime(net.simTime() + dist / PACKET_SPEED);
				net.insertEvent(msg);
				sent = true;
			}
		}

		return sent;

	}

	/**
	 * set a timer event
	 * @param interval the interval of this node
	 * @param index the type of the timer
	 *
	 */
	public void setTimer(float interval, Object index) {
		SimEventTimer timer =
			new SimEventTimer(net.simTime() + interval, this, index);
		//added by Guimin
		timer.setSender(this);
		timer.setRecipient(this);
		net.insertEvent(timer);
	}
	/** set a timer at micro-second level precision */
	public void setMSTimer(long ms_interval, Object index) {
		SimEventTimer timer =
			new SimEventTimer(net.getMSSimTime() + ms_interval, this, index);
		//added by Guimin
		timer.setSender(this);
		timer.setRecipient(this);
		net.insertEvent(timer);
	}

	/** add by Guangyu
	 * set a MAC timer event
	 * @param interval the interval of this node
	 * @param index the type of the timer
	 *
	 */
	public SimEventMAC setMACTimer(float interval, int index) {
		SimEventMAC timer =
			new SimEventMAC(net.simTime() + interval, this, index);
		timer.setSender(this);
		timer.setRecipient(this);
		net.insertEvent(timer);

		return timer;
	}
	public SimEventMAC setMSMACTimer(long ms_interval, int index) {
		SimEventMAC timer = new SimEventMAC(0, this, index);
		timer.setMSTime(net.getMSSimTime() + ms_interval);
		timer.setSender(this);
		timer.setRecipient(this);
		net.insertEvent(timer);

		return timer;
	}

	/**
	 * @return the simulation time
	 */
	public float simTime() {
		return net.simTime();
	}

	//message handling from/to Simulation Event Processor
	/**
	 * handle the events from simulation event
	 */
	public void receiveEvent(SimEvent event) {
		try{
			if ((event.getRecipient() != this) && (event.getRecipient() != null)) {
				lotusDebug.errorln("event dispatch error");
				return;
			}
			if (event.eventType() == SimEvent.PROTOCOL_TIMER_EVENT) {
				if (protocol == null) {
					lotusDebug.errorln("pending timer event is discarded");
					return;
				}
				protocol.handleTimer(((SimEventTimer)event).getIndex());
			} else if (event.eventType() == SimEvent.MAC_TIMER_EVENT) {
				//MAC timer
				//lotusDebug.errorln("MAC timer event");
				wl_MAC.handleMACTimer((SimEventMAC)event);
	
			} else if (event.eventType() == SimEvent.MESSAGE_EVENT) {
	
				//lotusDebug.errorln(this.toString()+"Receive MSG: "+event);
				net.sum_sent++;
	
				SimEventMsg msg = (SimEventMsg)event;
	
				if (msg.isWireless() && wl_lastMsgCollision) {
					lotusDebug.errorln(
						"collision. message dropped. time=" + net.simTime());
					net.sum_collision++;
					return;
				}
	
				if (!(net.getPhyNodes().contains(msg.getSrc()))) {
					lotusDebug.errorln("unknown sender/receiver packet");
					return;
				}
				if (msg.getDst() == this || msg.getDst() == null) {
					
					if(msg.getTTL()>0)msg.setTTL(msg.getTTL()-1);
					
					if (forwarding && msg.getDst() == null && msg.getTTL()!=0) {
						
						//lotusDebug.errorln(""+this+"forwarding multicast");
						
						//only forward multicast message to wired neighbors
						//Vector phyneighbors = this.getAllNeighbors();
						Vector phyneighbors = this.getWiredNeighbors();
						//lotusDebug.errorln("neighbor size="+this.getPhyNeighborNumber());
						
						for (int i = 0; i < phyneighbors.size(); i++) {
							
							OverlayNode pn = (OverlayNode)phyneighbors.elementAt(i);
							OverlayNode nextHop = pn.getPhyNextHop(msg.getSrc());
							
							//lotusDebug.errorln("test neighbor "+pn);
							
							
							if (nextHop == this) {
								//forward the multicast data
								SimEventMsg newmsg =(SimEventMsg)msg.clone();
								
								//lotusDebug.errorln(this.toString()+"forwarding multicast from "+msg.getSender()+" to "+pn);
								if (this.isWiredNeighbor(pn))
									newmsg.setTransType(SimEventMsg.TRANS_WIRED);
								else
									newmsg.setTransType(SimEventMsg.TRANS_WIRELESS);
	
								newmsg.setSender(this);
								newmsg.setRecipient(pn);
								newmsg.setTTL(msg.getTTL());
	
								if (this.isWiredNeighbor(pn)) {
									newmsg.setTime(
										net.simTime()
											+ getDirectDistanceTo(pn)
												/ PACKET_SPEED);
									net.insertEvent(newmsg);
	
								} else {
									newmsg.setTime(
										net.simTime()
											+ getDirectDistanceTo(pn)
												/ WirelessPhysicalLayer.PACKET_SPEED);
									wl_MAC.sendMsg(newmsg);
	
								}
							}
						}
					}
					if (protocol != null) {
						protocol.handleMsg(msg);
					}
	
				} else {
					//lotusDebug.errorln("receive message from "+msg.getSrc()+" to "+msg.getDst());
					if(msg.getTTL()>0)msg.setTTL(msg.getTTL()-1);
					if (!forwarding || msg.getTTL()==0)
						return;
	
					OverlayNode nextHop = getPhyNextHop(msg.getDst());
					if (null == nextHop) {
						lotusDebug.errorln("packet can not be routed");
						return;
					} else {
						if (this.isWiredNeighbor(nextHop))
							msg.setTransType(SimEventMsg.TRANS_WIRED);
						else
							msg.setTransType(SimEventMsg.TRANS_WIRELESS);
	
						//lotusDebug.errorln(this.toString()+"forwarding from "+msg.getSender()+" to "+nextHop);
						msg.setSender(this);
						msg.setRecipient(nextHop);
	
						if (this.isWiredNeighbor(nextHop)) {
							msg.setTime(
								net.simTime()
									+ getDirectDistanceTo(nextHop) / PACKET_SPEED);
							net.insertEvent(msg);
	
						} else {
							msg.setTime(
								net.simTime()
									+ getDirectDistanceTo(nextHop)
										/ WirelessPhysicalLayer.PACKET_SPEED);
							wl_MAC.sendMsg(msg);
	
						}
					}
				}
			}
		}catch(Exception e){
			lotusDebug.println("Exception when handling event: " + event);
			e.printStackTrace();
		}
	}

	/** set the logical node running on this node */
	void setProtocol(protocolInterface p) {
		protocol = p;
	}

	/** return the logical node running on this node*/
	protocolInterface getProtocol() {
		return protocol;
	}

	public String toString() {
		if (getProtocol() == null)
			return "@" + addr;
		else
			return getProtocol().toString();
	}
	/* (non-Javadoc)
	 * @see physicalInterface#getTimer(int)
	 */
	public float getTimer(Object index) {
		return net.getEvent(new SimEventTimer(0, this, index));
	}
	/* (non-Javadoc)
	 * @see physicalInterface#clearTimer(int)
	 */
	public void clearTimer(Object index) {
		net.clearEvent(new SimEventTimer(0, this, index));
	}
	public boolean equals(Object o){
		return this == o;
	}
}
